# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""Typing for half-precision (fp16/bf16) intrinsic functions.

Registers bf16 intrinsics dynamically at typing context initialization time
to handle module identity issues with the numba_cuda_mlir redirector.
"""

from numba_cuda_mlir.numba_cuda.typing.templates import (
    ConcreteTemplate,
    Registry,
    signature,
)
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.types.ext_types import bfloat16 as bf16

registry = Registry()
_bf16_registered_globals = set()

# List of bf16 intrinsic names - only unary ops exist in numba-cuda's cuda_bf16
_BF16_UNARY_INTRINSICS = [
    "htrunc",
    "hceil",
    "hfloor",
    "hrint",
    "hsqrt",
    "hrsqrt",
    "hrcp",
    "hlog",
    "hlog2",
    "hlog10",
    "hcos",
    "hsin",
    "hexp",
    "hexp2",
    "hexp10",
    "htanh",
    "htanh_approx",
]


def register_bf16_globals():
    """Register bf16 intrinsics as globals. Must be called after bf16 module is imported."""
    import sys

    mod = sys.modules.get("numba_cuda_mlir.numba_cuda._internal.cuda_bf16")
    if not mod:
        return

    # Create and register templates for unary intrinsics
    for name in _BF16_UNARY_INTRINSICS:
        func = getattr(mod, name, None)
        if func and func not in _bf16_registered_globals:

            class UnaryTemplate(ConcreteTemplate):
                key = func
                cases = [signature(bf16, bf16)]

            UnaryTemplate.__name__ = f"BF16_{name}_Template"
            registry.register(UnaryTemplate)
            registry.register_global(func, types.Function(UnaryTemplate))
            _bf16_registered_globals.add(func)

    # Register typing for bitcast functions that also accept integer types
    # This allows passing integer bit patterns to be treated as bf16
    _bfloat16_as_short = getattr(mod, "__bfloat16_as_short", None)
    if _bfloat16_as_short and _bfloat16_as_short not in _bf16_registered_globals:
        _key = _bfloat16_as_short

        class BF16AsShortTemplate(ConcreteTemplate):
            key = _key
            cases = [
                signature(types.int16, bf16),
                signature(types.int16, types.int64),
                signature(types.int16, types.int32),
                signature(types.int16, types.int16),
            ]

        registry.register(BF16AsShortTemplate)
        registry.register_global(_bfloat16_as_short, types.Function(BF16AsShortTemplate))
        _bf16_registered_globals.add(_bfloat16_as_short)

    _bfloat16_as_ushort = getattr(mod, "__bfloat16_as_ushort", None)
    if _bfloat16_as_ushort and _bfloat16_as_ushort not in _bf16_registered_globals:
        _key = _bfloat16_as_ushort

        class BF16AsUShortTemplate(ConcreteTemplate):
            key = _key
            cases = [
                signature(types.uint16, bf16),
                signature(types.uint16, types.int64),
                signature(types.uint16, types.int32),
                signature(types.uint16, types.int16),
                signature(types.uint16, types.uint64),
                signature(types.uint16, types.uint32),
                signature(types.uint16, types.uint16),
            ]

        registry.register(BF16AsUShortTemplate)
        registry.register_global(_bfloat16_as_ushort, types.Function(BF16AsUShortTemplate))
        _bf16_registered_globals.add(_bfloat16_as_ushort)

    _short_as_bfloat16 = getattr(mod, "__short_as_bfloat16", None)
    if _short_as_bfloat16 and _short_as_bfloat16 not in _bf16_registered_globals:
        _key = _short_as_bfloat16

        class ShortAsBF16Template(ConcreteTemplate):
            key = _key
            cases = [
                signature(bf16, types.int16),
                signature(bf16, types.int64),
                signature(bf16, types.int32),
            ]

        registry.register(ShortAsBF16Template)
        registry.register_global(_short_as_bfloat16, types.Function(ShortAsBF16Template))
        _bf16_registered_globals.add(_short_as_bfloat16)

    _ushort_as_bfloat16 = getattr(mod, "__ushort_as_bfloat16", None)
    if _ushort_as_bfloat16 and _ushort_as_bfloat16 not in _bf16_registered_globals:
        _key = _ushort_as_bfloat16

        class UShortAsBF16Template(ConcreteTemplate):
            key = _key
            cases = [
                signature(bf16, types.uint16),
                signature(bf16, types.int64),
                signature(bf16, types.int32),
                signature(bf16, types.uint64),
                signature(bf16, types.uint32),
            ]

        registry.register(UShortAsBF16Template)
        registry.register_global(_ushort_as_bfloat16, types.Function(UShortAsBF16Template))
        _bf16_registered_globals.add(_ushort_as_bfloat16)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Thunk that overrides everything in numba.cuda and overrides
everything in numba.cuda that numba_cuda_mlir _also_ supports.
"""

import importlib
import sys

# Base API comes from `numba.cuda` (which may be redirected to numba-cuda-mlir).
from numba_cuda_mlir.numba_cuda import *  # noqa: F403
from numba_cuda_mlir.numba_cuda.cudadrv.devicearray import (
    DeviceNDArray,  # ty:ignore[unresolved-import]
)  # noqa: F401
from numba_cuda_mlir.numba_cuda.misc.special import literal_unroll  # noqa: F401,E402
from numba_cuda_mlir.numba_cuda.np.numpy_support import carray, farray  # noqa: F401

from numba_cuda_mlir.decorators import mlir_jit as jit
from numba_cuda_mlir.compiler import (
    compile,
    compile_ptx,
    compile_all,
    compile_mlir,
    compile_for,
    compile_cubin,
    declare_device,
)


HAS_NUMBA = False


# Submodules (must be modules, not class stubs) so that `numba.cuda.shared` and
# `cuda.shared.array` resolve to the same callables we register typing/lowering
# for.  Assign from importlib's return value so the star import from
# numba_cuda_mlir.numba_cuda cannot leave stub attributes on this package.
cudadrv = importlib.import_module("numba_cuda_mlir.cuda.cudadrv")
const = importlib.import_module("numba_cuda_mlir.cuda.const")
local = importlib.import_module("numba_cuda_mlir.cuda.local")
shared = importlib.import_module("numba_cuda_mlir.cuda.shared")
fp16 = importlib.import_module("numba_cuda_mlir.cuda.fp16")
libdevice = importlib.import_module("numba_cuda_mlir.cuda.libdevice")
libdevicefuncs = importlib.import_module("numba_cuda_mlir.cuda.libdevicefuncs")
vector = importlib.import_module("numba_cuda_mlir.cuda.vector")
vector_types = importlib.import_module("numba_cuda_mlir.cuda.vector_types")

# Expose vector type constructors (float32x4, int32x2, etc.) at module level
from .vector_types import *  # noqa: F401,F403

local_array = local.array  # noqa: F401
shared_array = shared.array  # noqa: F401


def __getattr__(name):
    """Lazy load modules to avoid circular import issues."""
    if name in ("intrin", "tensor_map", "experimental"):
        import importlib

        module = importlib.import_module(f"numba_cuda_mlir.cuda.{name}")
        globals()[name] = module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def inline_ptx(format_string: str, *args) -> None:
    """
    Add PTX code directly into the kernel.
    The format string and arguments mirror the CUDA C++ inline assembly syntax.
    """


def vectorize(*args, **kwargs):
    raise NotImplementedError("vectorize is not implemented")


def clz(x):
    """Count leading zeros. For a 32-bit value, returns 0-32. For 64-bit, returns 0-64."""
    pass


def ffs(x):
    """Find first set bit. Returns 1-indexed position of LSB, or 0 if input is 0."""
    pass


def brev(x):
    """Reverse the bits of x."""
    pass


def popc(x):
    """Count the number of set bits in x."""
    pass


def selp(cond, a, b):
    """Select based on predicate: returns a if cond is true, else b."""
    pass


# Special registers - these are accessed as module attributes, not functions
warpsize = 32
laneid = None  # Placeholder - actual value comes from NVVM intrinsic at runtime


# Cache hint load instructions
def ldca(array, i):
    """Generate a `ld.global.ca` instruction for element `i` of an array."""
    pass


def ldcg(array, i):
    """Generate a `ld.global.cg` instruction for element `i` of an array."""
    pass


def ldcs(array, i):
    """Generate a `ld.global.cs` instruction for element `i` of an array."""
    pass


def ldlu(array, i):
    """Generate a `ld.global.lu` instruction for element `i` of an array."""
    pass


def ldcv(array, i):
    """Generate a `ld.global.cv` instruction for element `i` of an array."""
    pass


# Cache hint store instructions
def stcg(array, i, value):
    """Generate a `st.global.cg` instruction for element `i` of an array."""
    pass


def stcs(array, i, value):
    """Generate a `st.global.cs` instruction for element `i` of an array."""
    pass


def stwb(array, i, value):
    """Generate a `st.global.wb` instruction for element `i` of an array."""
    pass


def stwt(array, i, value):
    """Generate a `st.global.wt` instruction for element `i` of an array."""
    pass



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.lowering_utilities import int_of
from numba_cuda_mlir.lowering_utilities import convert_tuple_like
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir.errors import InternalCompilerError
from numba_cuda_mlir.mlir_lowering import MLIRLower
from textwrap import dedent
from numba_cuda_mlir.lowering_utilities import convert, tensor_to_memref
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda import typing as typing
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir._mlir.dialects import func, llvm, tensor, memref
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower
from numba_cuda_mlir.numba_cuda.extending import overload, intrinsic
import numba_cuda_mlir.cuda.libdevice as libdevice
import numba_cuda_mlir.cuda.libdevicefuncs as libdevicefuncs
import inspect
import ast
from numba_cuda_mlir.lowering_utilities import get_or_insert_function
from numba_cuda_mlir.numba_cuda.typing.templates import ConcreteTemplate
from typing import Callable, Any

_libdevice_descriptors: dict[Callable[..., Any], libdevicefuncs.Descriptor] = {}


def _generalize_type(ty: types.Type) -> types.Type:
    """
    Generalize types.int64 -> types.Integer(), recursively.
    """
    match ty:
        case types.UniTuple() as ut:
            return types.UniTuple(_generalize_type(ut.dtype), ut.count)
        case types.Tuple() as t:
            return types.Tuple(_generalize_type(ty) for ty in t.types)
        case types.Integer():
            return types.Integer
        case types.Float():
            return types.Float
        case _:
            raise NotImplementedError(f"Unknown type: {ty}")


def libdevice_implement_by_pointer(pyfunc, py_api: typing.Signature, c_api: typing.Signature):
    def core(lower: MLIRLower, target, args, kwargs):
        trace("lower: %s", pyfunc.__name__)
        args = [lower.load_var(arg) for arg in args]
        by_ptr_return_types = [ty for ty in c_api.args if isinstance(ty, types.CPointer)]
        by_ptr_return_mlir_types = [lower.get_mlir_type(ty.dtype) for ty in by_ptr_return_types]
        by_ptr_return_vars = [lower.alloca(mlir_type) for mlir_type in by_ptr_return_mlir_types]
        args += by_ptr_return_vars
        descriptor = _libdevice_descriptors[pyfunc]
        fn_type: ir.FunctionType = lower.get_mlir_type(descriptor.c_sig)
        libdevice_name = "__nv_" + pyfunc.__name__
        callee = get_or_insert_function(libdevice_name, fn_type, lower.mlir_gpu_module)
        result = [] if c_api.return_type is types.void else [fn_type.results[0]]
        call = func.call(result=result, callee=callee.name.value, operands_=args)
        result = [] if c_api.return_type is types.void else [call]
        result += [
            llvm.load(ty, var) for ty, var in zip(by_ptr_return_mlir_types, by_ptr_return_vars)
        ]
        lower.store_var(target, tuple(result))

    core.__name__ = f"lower_{pyfunc.__name__}_by_pointer"
    py_args = tuple(_generalize_type(ty) for ty in py_api.args)
    lower(pyfunc, *py_args)(core)


def libdevice_implement_uniform(pyfunc, api):
    def core(lower, target, args, kwargs):
        args = [lower.load_var(arg) for arg in args]
        trace("%s", pyfunc.__name__)
        descriptor = _libdevice_descriptors[pyfunc]
        fn_type: ir.FunctionType = lower.get_mlir_type(descriptor.c_sig)
        libdevice_name = "__nv_" + pyfunc.__name__
        callee = get_or_insert_function(libdevice_name, fn_type, lower.mlir_gpu_module)
        args = [convert(arg, ty) for arg, ty in zip(args, fn_type.inputs)]
        call = func.call(result=[fn_type.results[0]], callee=callee.name.value, operands_=args)
        call = convert(call, lower.get_mlir_type(target))
        lower.store_var(target, call)

    core.__name__ = f"lower_{pyfunc.__name__}_uniform"
    arg_types = tuple(_generalize_type(ty) for ty in api.args)
    lower(pyfunc, *arg_types)(core)


def libdevice_implement(pyfunc, py_api, c_api):
    if py_api == c_api:
        libdevice_implement_uniform(pyfunc, py_api)
    else:
        libdevice_implement_by_pointer(pyfunc, py_api, c_api)


def _libdevice_register():
    """
    Register type declarations and lowerings for all libdevice functions.
    """
    for descriptor in libdevicefuncs.libdevice_descriptors():
        py_sig = descriptor.py_sig
        c_sig = descriptor.c_sig
        pyfunc = getattr(libdevice, descriptor.py_name)
        _libdevice_descriptors[pyfunc] = descriptor
        libdevice_implement(pyfunc, py_sig, c_sig)


_libdevice_register()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause


class prange:
    """Provides a 1D parallel iterator that generates a sequence of integers.
    In non-parallel contexts, prange is identical to range.
    """

    def __new__(cls, *args):
        return range(*args)


def literally(obj):
    """Forces Numba to interpret *obj* as an Literal value.

    *obj* must be either a literal or an argument of the caller function, where
    the argument must be bound to a literal. The literal requirement
    propagates up the call stack.

    This function is intercepted by the compiler to alter the compilation
    behavior to wrap the corresponding function parameters as ``Literal``.
    It has **no effect** outside of nopython-mode (interpreter, and objectmode).

    The current implementation detects literal arguments in two ways:

    1. Scans for uses of ``literally`` via a compiler pass.
    2. ``literally`` is overloaded to raise ``numba.errors.ForceLiteralArg``
       to signal the dispatcher to treat the corresponding parameter
       differently. This mode is to support indirect use (via a function call).

    The execution semantic of this function is equivalent to an identity
    function.

    See :ghfile:`numba/tests/test_literal_dispatch.py` for examples.
    """
    return obj


def literal_unroll(container):
    return container


__all__ = [
    "prange",
    "literally",
    "literal_unroll",
]



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import (
    List as _List,
    Optional as _Optional,
    Sequence as _Sequence,
    Tuple as _Tuple,
    Type as _Type,
    Union as _Union,
)

from .._mlir_libs import _mlir as _cext
from ..ir import (
    ArrayAttr,
    Attribute,
    BoolAttr,
    DenseI64ArrayAttr,
    IntegerAttr,
    IntegerType,
    OpView,
    Operation,
    ShapedType,
    Value,
)

__all__ = [
    "equally_sized_accessor",
    "get_default_loc_context",
    "get_op_result_or_value",
    "get_op_results_or_values",
    "get_op_result_or_op_results",
    "segmented_accessor",
]


def segmented_accessor(elements, raw_segments, idx):
    """
    Returns a slice of elements corresponding to the idx-th segment.

      elements: a sliceable container (operands or results).
      raw_segments: an mlir.ir.Attribute, of DenseI32Array subclass containing
          sizes of the segments.
      idx: index of the segment.
    """
    segments = _cext.ir.DenseI32ArrayAttr(raw_segments)
    start = sum(segments[i] for i in range(idx))
    end = start + segments[idx]
    return elements[start:end]


def equally_sized_accessor(
    elements, n_simple, n_variadic, n_preceding_simple, n_preceding_variadic
):
    """
    Returns a starting position and a number of elements per variadic group
    assuming equally-sized groups and the given numbers of preceding groups.

      elements: a sequential container.
      n_simple: the number of non-variadic groups in the container.
      n_variadic: the number of variadic groups in the container.
      n_preceding_simple: the number of non-variadic groups preceding the current
          group.
      n_preceding_variadic: the number of variadic groups preceding the current
          group.
    """

    total_variadic_length = len(elements) - n_simple
    # This should be enforced by the C++-side trait verifier.
    assert total_variadic_length % n_variadic == 0

    elements_per_group = total_variadic_length // n_variadic
    start = n_preceding_simple + n_preceding_variadic * elements_per_group
    return start, elements_per_group


def get_default_loc_context(location=None):
    """
    Returns a context in which the defaulted location is created. If the location
    is None, takes the current location from the stack.
    """
    if location is None:
        if _cext.ir.Location.current:
            return _cext.ir.Location.current.context
        return None
    return location.context


def get_op_result_or_value(
    arg: _Union[
        _cext.ir.OpView, _cext.ir.Operation, _cext.ir.Value, _cext.ir.OpResultList
    ]
) -> _cext.ir.Value:
    """Returns the given value or the single result of the given op.

    This is useful to implement op constructors so that they can take other ops as
    arguments instead of requiring the caller to extract results for every op.
    Raises ValueError if provided with an op that doesn't have a single result.
    """
    if isinstance(arg, _cext.ir.OpView):
        return arg.operation.result
    elif isinstance(arg, _cext.ir.Operation):
        return arg.result
    elif isinstance(arg, _cext.ir.OpResultList):
        return arg[0]
    else:
        assert isinstance(arg, _cext.ir.Value), f"expects Value, got {type(arg)}"
        return arg


def get_op_results_or_values(
    arg: _Union[
        _cext.ir.OpView,
        _cext.ir.Operation,
        _Sequence[_Union[_cext.ir.OpView, _cext.ir.Operation, _cext.ir.Value]],
    ]
) -> _Union[
    _Sequence[_Union[_cext.ir.OpView, _cext.ir.Operation, _cext.ir.Value]],
    _cext.ir.OpResultList,
]:
    """Returns the given sequence of values or the results of the given op.

    This is useful to implement op constructors so that they can take other ops as
    lists of arguments instead of requiring the caller to extract results for
    every op.
    """
    if isinstance(arg, _cext.ir.OpView):
        return arg.operation.results
    elif isinstance(arg, _cext.ir.Operation):
        return arg.results
    else:
        return arg


def get_op_result_or_op_results(
    op: _Union[_cext.ir.OpView, _cext.ir.Operation],
) -> _Union[_cext.ir.Operation, _cext.ir.OpResult, _Sequence[_cext.ir.OpResult]]:
    results = op.results
    num_results = len(results)
    if num_results == 1:
        return results[0]
    elif num_results > 1:
        return results
    elif isinstance(op, _cext.ir.OpView):
        return op.operation
    else:
        return op


ResultValueTypeTuple = _cext.ir.Operation, _cext.ir.OpView, _cext.ir.Value
ResultValueT = _Union[ResultValueTypeTuple]
VariadicResultValueT = _Union[ResultValueT, _Sequence[ResultValueT]]

StaticIntLike = _Union[int, IntegerAttr]
ValueLike = _Union[Operation, OpView, Value]
MixedInt = _Union[StaticIntLike, ValueLike]

IntOrAttrList = _Sequence[_Union[IntegerAttr, int]]
OptionalIntList = _Optional[_Union[ArrayAttr, IntOrAttrList]]

BoolOrAttrList = _Sequence[_Union[BoolAttr, bool]]
OptionalBoolList = _Optional[_Union[ArrayAttr, BoolOrAttrList]]

MixedValues = _Union[_Sequence[_Union[StaticIntLike, ValueLike]], ArrayAttr, ValueLike]

DynamicIndexList = _Sequence[_Union[MixedInt, _Sequence[MixedInt]]]


def _dispatch_dynamic_index_list(
    indices: _Union[DynamicIndexList, ArrayAttr],
) -> _Tuple[_List[ValueLike], _Union[_List[int], ArrayAttr], _List[bool]]:
    """Dispatches a list of indices to the appropriate form.

    This is similar to the custom `DynamicIndexList` directive upstream:
    provided indices may be in the form of dynamic SSA values or static values,
    and they may be scalable (i.e., as a singleton list) or not. This function
    dispatches each index into its respective form. It also extracts the SSA
    values and static indices from various similar structures, respectively.
    """
    dynamic_indices = []
    static_indices = [ShapedType.get_dynamic_size()] * len(indices)
    scalable_indices = [False] * len(indices)

    # ArrayAttr: Extract index values.
    if isinstance(indices, ArrayAttr):
        indices = [idx for idx in indices]

    def process_nonscalable_index(i, index):
        """Processes any form of non-scalable index.

        Returns False if the given index was scalable and thus remains
        unprocessed; True otherwise.
        """
        if isinstance(index, int):
            static_indices[i] = index
        elif isinstance(index, IntegerAttr):
            static_indices[i] = index.value  # pytype: disable=attribute-error
        elif isinstance(index, (Operation, Value, OpView)):
            dynamic_indices.append(index)
        else:
            return False
        return True

    # Process each index at a time.
    for i, index in enumerate(indices):
        if not process_nonscalable_index(i, index):
            # If it wasn't processed, it must be a scalable index, which is
            # provided as a _Sequence of one value, so extract and process that.
            scalable_indices[i] = True
            assert len(index) == 1
            ret = process_nonscalable_index(i, index[0])
            assert ret

    return dynamic_indices, static_indices, scalable_indices


# Dispatches `MixedValues` that all represents integers in various forms into
# the following three categories:
#   - `dynamic_values`: a list of `Value`s, potentially from op results;
#   - `packed_values`: a value handle, potentially from an op result, associated
#                      to one or more payload operations of integer type;
#   - `static_values`: an `ArrayAttr` of `i64`s with static values, from Python
#                      `int`s, from `IntegerAttr`s, or from an `ArrayAttr`.
# The input is in the form for `packed_values`, only that result is set and the
# other two are empty. Otherwise, the input can be a mix of the other two forms,
# and for each dynamic value, a special value is added to the `static_values`.
def _dispatch_mixed_values(
    values: MixedValues,
) -> _Tuple[_List[Value], _Union[Operation, Value, OpView], DenseI64ArrayAttr]:
    dynamic_values = []
    packed_values = None
    static_values = None
    if isinstance(values, ArrayAttr):
        static_values = values
    elif isinstance(values, (Operation, Value, OpView)):
        packed_values = values
    else:
        static_values = []
        for size in values or []:
            if isinstance(size, int):
                static_values.append(size)
            elif isinstance(size, IntegerAttr):
                static_values.append(size.value)
            else:
                static_values.append(ShapedType.get_dynamic_size())
                dynamic_values.append(size)
        static_values = DenseI64ArrayAttr.get(static_values)

    return (dynamic_values, packed_values, static_values)


def _get_value_or_attribute_value(
    value_or_attr: _Union[any, Attribute, ArrayAttr]
) -> any:
    if isinstance(value_or_attr, Attribute) and hasattr(value_or_attr, "value"):
        return value_or_attr.value
    if isinstance(value_or_attr, ArrayAttr):
        return _get_value_list(value_or_attr)
    return value_or_attr


def _get_value_list(
    sequence_or_array_attr: _Union[_Sequence[any], ArrayAttr]
) -> _Sequence[any]:
    return [_get_value_or_attribute_value(v) for v in sequence_or_array_attr]


def _get_int_array_attr(
    values: _Optional[_Union[ArrayAttr, IntOrAttrList]]
) -> ArrayAttr:
    if values is None:
        return None

    # Turn into a Python list of Python ints.
    values = _get_value_list(values)

    # Make an ArrayAttr of IntegerAttrs out of it.
    return ArrayAttr.get(
        [IntegerAttr.get(IntegerType.get_signless(64), v) for v in values]
    )


def _get_int_array_array_attr(
    values: _Optional[_Union[ArrayAttr, _Sequence[_Union[ArrayAttr, IntOrAttrList]]]]
) -> ArrayAttr:
    """Creates an ArrayAttr of ArrayAttrs of IntegerAttrs.

    The input has to be a collection of a collection of integers, where any
    Python _Sequence and ArrayAttr are admissible collections and Python ints and
    any IntegerAttr are admissible integers. Both levels of collections are
    turned into ArrayAttr; the inner level is turned into IntegerAttrs of i64s.
    If the input is None, an empty ArrayAttr is returned.
    """
    if values is None:
        return None

    # Make sure the outer level is a list.
    values = _get_value_list(values)

    # The inner level is now either invalid or a mixed sequence of ArrayAttrs and
    # Sequences. Make sure the nested values are all lists.
    values = [_get_value_list(nested) for nested in values]

    # Turn each nested list into an ArrayAttr.
    values = [_get_int_array_attr(nested) for nested in values]

    # Turn the outer list into an ArrayAttr.
    return ArrayAttr.get(values)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
MLIR-based NRT (Numba Runtime) lowering for numba_cuda_mlir.

Provides MLIR LLVM dialect equivalents of numba-cuda's NRT intrinsics,
enabling on-device heap allocation for strings, arrays, etc.
"""

import sys

from numba_cuda_mlir import types
from numba_cuda_mlir.numba_cuda.cpython.unicode import (
    _malloc_string,
    deref_uint8,
    deref_uint16,
    deref_uint32,
    set_uint8,
    set_uint16,
    set_uint32,
)
from numba_cuda_mlir._mlir.dialects import func
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir.mlir.dialect_exts import llvm

from numba_cuda_mlir.lowering_utilities import (
    get_or_insert_function,
    convert,
    GEP_DYNAMIC_INDEX,
    int_of,
)
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower

_HASH_WIDTH = 64 if sys.maxsize > 2**32 else 32


def _get_or_declare(module, name, result_types, arg_types):
    func_type = ir.FunctionType.get(arg_types, result_types)
    return get_or_insert_function(name, func_type, module)


def nrt_meminfo_alloc(module, size, align=None):
    """Allocate a MemInfo. Uses NRT_MemInfo_alloc_aligned (the only CUDA variant)."""
    if align is None:
        align = int_of(8, T.i32())
    callee = _get_or_declare(module, "NRT_MemInfo_alloc_aligned", [llvm.ptr()], [T.i64(), T.i32()])
    return func.call(result=[llvm.ptr()], callee=callee.name.value, operands_=[size, align])


def nrt_meminfo_data(module, meminfo):
    callee = _get_or_declare(module, "NRT_MemInfo_data_fast", [llvm.ptr()], [llvm.ptr()])
    return func.call(result=[llvm.ptr()], callee=callee.name.value, operands_=[meminfo])


def nrt_incref(module, meminfo):
    callee = _get_or_declare(module, "NRT_incref", [], [llvm.ptr()])
    func.call(result=[], callee=callee.name.value, operands_=[meminfo])


def nrt_decref(module, meminfo):
    callee = _get_or_declare(module, "NRT_decref", [], [llvm.ptr()])
    func.call(result=[], callee=callee.name.value, operands_=[meminfo])


def _lower_malloc_string(builder, target, args, kwargs):
    """MLIR equivalent of numba-cuda's _malloc_string intrinsic."""
    kind_val, char_bytes_val, length_val, is_ascii_val = builder.load_vars(args)
    module = builder.mlir_gpu_module

    kind_val = convert(kind_val, T.i32())
    char_bytes_val = convert(char_bytes_val, T.i64())
    length_val = convert(length_val, T.i64())
    is_ascii_val = convert(is_ascii_val, T.i32())

    nbytes = char_bytes_val * (length_val + 1)

    meminfo = nrt_meminfo_alloc(module, nbytes)
    data = nrt_meminfo_data(module, meminfo)

    hash_type = ir.IntegerType.get_signless(_HASH_WIDTH)
    neg_one = int_of(-1, hash_type)
    null_ptr = llvm.mlir_zero(res=llvm.ptr())

    struct_type = llvm.StructType.get_literal(
        [llvm.ptr(), T.i64(), T.i32(), T.i32(), hash_type, llvm.ptr(), llvm.ptr()]
    )
    desc = llvm.mlir_undef(res=struct_type)
    for i, value in enumerate(
        [data, length_val, kind_val, is_ascii_val, neg_one, meminfo, null_ptr]
    ):
        desc = llvm.insertvalue(desc, value, i)

    builder.store_var(target, desc)


def _make_deref_lowering(bitsize):
    """MLIR lowering for deref_uint{8,16,32}: read a character from string data."""

    def _lower_deref(builder, target, args, kwargs):
        data, idx = builder.load_vars(args)
        data = convert(data, llvm.ptr())
        idx = convert(idx, T.i64())
        int_type = ir.IntegerType.get_signless(bitsize)
        addr = llvm.getelementptr(llvm.ptr(), data, [idx], [GEP_DYNAMIC_INDEX], int_type, None)
        loaded = llvm.load(int_type, addr)
        result = convert(loaded, T.i32())
        builder.store_var(target, result)

    return _lower_deref


def _make_set_lowering(bitsize):
    """MLIR lowering for set_uint{8,16,32}: write a character to string data."""

    def _lower_set(builder, target, args, kwargs):
        data, idx, ch = builder.load_vars(args)
        data = convert(data, llvm.ptr())
        idx = convert(idx, T.i64())
        ch = convert(ch, T.i32())
        int_type = ir.IntegerType.get_signless(bitsize)
        ch = convert(ch, int_type)
        addr = llvm.getelementptr(llvm.ptr(), data, [idx], [GEP_DYNAMIC_INDEX], int_type, None)
        llvm.store(ch, addr)

    return _lower_set


lower(_malloc_string, types.Any, types.Any, types.Any, types.Any)(_lower_malloc_string)

lower(deref_uint8, types.Any, types.Any)(_make_deref_lowering(8))
lower(deref_uint16, types.Any, types.Any)(_make_deref_lowering(16))
lower(deref_uint32, types.Any, types.Any)(_make_deref_lowering(32))

lower(set_uint8, types.Any, types.Any, types.Any)(_make_set_lowering(8))
lower(set_uint16, types.Any, types.Any, types.Any)(_make_set_lowering(16))
lower(set_uint32, types.Any, types.Any, types.Any)(_make_set_lowering(32))



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Implementation of operations on numpy timedelta64.
"""

import numpy as np
import operator

from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda import cgutils
from numba_cuda_mlir.numba_cuda.cgutils import create_constant_array
from numba_cuda_mlir.numba_cuda.core.imputils import (
    impl_ret_untracked,
    lower_cast,
    Registry,
)
from numba_cuda_mlir.numba_cuda.np import npdatetime_helpers, numpy_support
from numba_cuda_mlir.numba_cuda.extending import overload_method
from numba_cuda_mlir.numba_cuda.core.config import IS_32BITS
from numba_cuda_mlir.numba_cuda.core.errors import LoweringError

# datetime64 and timedelta64 use the same internal representation;
# llvmlite type/constant discarded on the MLIR path (used only by dead codegen)
DATETIME64 = TIMEDELTA64 = None
NAT = None

TIMEDELTA_BINOP_SIG = (types.NPTimedelta,) * 2

registry = Registry("np.npdatetime")
lower = registry.lower
lower_constant = registry.lower_constant

_DEAD_CODEGEN_MSG = "this datetime llvmlite codegen is not used on the MLIR path"


# The datetime/timedelta codegen helpers and @lower ops below built llvmlite IR
# (Constant / llvmlite.ir). They are filtered out on the MLIR path (datetime is
# lowered by numba_cuda_mlir.lowering.datetime), so they are dead stubs.
def scale_by_constant(builder, val, factor):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


def unscale_by_constant(builder, val, factor):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


def add_constant(builder, val, const):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


def scale_timedelta(context, builder, val, srcty, destty):
    """
    Scale the timedelta64 *val* from *srcty* to *destty*
    (both numba.types.NPTimedelta instances)
    """
    factor = npdatetime_helpers.get_timedelta_conversion_factor(srcty.unit, destty.unit)
    if factor is None:
        # This can happen when using explicit output in a ufunc.
        msg = f"cannot convert timedelta64 from {srcty.unit} to {destty.unit}"
        raise LoweringError(msg)
    return scale_by_constant(builder, val, factor)


def normalize_timedeltas(context, builder, left, right, leftty, rightty):
    """
    Scale either *left* or *right* to the other's unit, in order to have
    homogeneous units.
    """
    factor = npdatetime_helpers.get_timedelta_conversion_factor(leftty.unit, rightty.unit)
    if factor is not None:
        return scale_by_constant(builder, left, factor), right
    factor = npdatetime_helpers.get_timedelta_conversion_factor(rightty.unit, leftty.unit)
    if factor is not None:
        return left, scale_by_constant(builder, right, factor)
    # Typing should not let this happen, except on == and != operators
    raise RuntimeError("cannot normalize %r and %r" % (leftty, rightty))


def alloc_timedelta_result(builder, name="ret"):
    """
    Allocate a NaT-initialized datetime64 (or timedelta64) result slot.
    """
    ret = cgutils.alloca_once(builder, TIMEDELTA64, name=name)
    builder.store(NAT, ret)
    return ret


def alloc_boolean_result(builder, name="ret"):
    """
    Allocate an uninitialized boolean result slot.
    """
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


def is_not_nat(builder, val):
    """
    Return a predicate which is true if *val* is not NaT.
    """
    return builder.icmp_unsigned("!=", val, NAT)


def are_not_nat(builder, vals):
    """
    Return a predicate which is true if all of *vals* are not NaT.
    """
    assert len(vals) >= 1
    pred = is_not_nat(builder, vals[0])
    for val in vals[1:]:
        pred = builder.and_(pred, is_not_nat(builder, val))
    return pred


# llvmlite constant arrays discarded on the MLIR path (used only by dead codegen)
normal_year_months = None
leap_year_months = None
normal_year_months_acc = None
leap_year_months_acc = None


@lower_constant(types.NPDatetime)
@lower_constant(types.NPTimedelta)
def datetime_constant(context, builder, ty, pyval):
    return DATETIME64(pyval.astype(np.int64))


# Arithmetic operators on timedelta64


@lower(operator.pos, types.NPTimedelta)
def timedelta_pos_impl(context, builder, sig, args):
    res = args[0]
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(operator.neg, types.NPTimedelta)
def timedelta_neg_impl(context, builder, sig, args):
    res = builder.neg(args[0])
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(abs, types.NPTimedelta)
def timedelta_abs_impl(context, builder, sig, args):
    (val,) = args
    ret = alloc_timedelta_result(builder)
    with builder.if_else(cgutils.is_scalar_neg(builder, val)) as (
        then,
        otherwise,
    ):
        with then:
            builder.store(builder.neg(val), ret)
        with otherwise:
            builder.store(val, ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


def timedelta_sign_impl(context, builder, sig, args):
    """
    np.sign(timedelta64)
    """
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


@lower(operator.add, *TIMEDELTA_BINOP_SIG)
@lower(operator.iadd, *TIMEDELTA_BINOP_SIG)
def timedelta_add_impl(context, builder, sig, args):
    [va, vb] = args
    [ta, tb] = sig.args
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, are_not_nat(builder, [va, vb])):
        va = scale_timedelta(context, builder, va, ta, sig.return_type)
        vb = scale_timedelta(context, builder, vb, tb, sig.return_type)
        builder.store(builder.add(va, vb), ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(operator.sub, *TIMEDELTA_BINOP_SIG)
@lower(operator.isub, *TIMEDELTA_BINOP_SIG)
def timedelta_sub_impl(context, builder, sig, args):
    [va, vb] = args
    [ta, tb] = sig.args
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, are_not_nat(builder, [va, vb])):
        va = scale_timedelta(context, builder, va, ta, sig.return_type)
        vb = scale_timedelta(context, builder, vb, tb, sig.return_type)
        builder.store(builder.sub(va, vb), ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


def _timedelta_times_number(
    context, builder, td_arg, td_type, number_arg, number_type, return_type
):
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, is_not_nat(builder, td_arg)):
        if isinstance(number_type, types.Float):
            val = builder.sitofp(td_arg, number_arg.type)
            val = builder.fmul(val, number_arg)
            val = _cast_to_timedelta(context, builder, val)
        else:
            val = builder.mul(td_arg, number_arg)
        # The scaling is required for ufunc np.multiply() with an explicit
        # output in a different unit.
        val = scale_timedelta(context, builder, val, td_type, return_type)
        builder.store(val, ret)
    return builder.load(ret)


@lower(operator.mul, types.NPTimedelta, types.Integer)
@lower(operator.imul, types.NPTimedelta, types.Integer)
@lower(operator.mul, types.NPTimedelta, types.Float)
@lower(operator.imul, types.NPTimedelta, types.Float)
def timedelta_times_number(context, builder, sig, args):
    res = _timedelta_times_number(
        context,
        builder,
        args[0],
        sig.args[0],
        args[1],
        sig.args[1],
        sig.return_type,
    )
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(operator.mul, types.Integer, types.NPTimedelta)
@lower(operator.imul, types.Integer, types.NPTimedelta)
@lower(operator.mul, types.Float, types.NPTimedelta)
@lower(operator.imul, types.Float, types.NPTimedelta)
def number_times_timedelta(context, builder, sig, args):
    res = _timedelta_times_number(
        context,
        builder,
        args[1],
        sig.args[1],
        args[0],
        sig.args[0],
        sig.return_type,
    )
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(operator.truediv, types.NPTimedelta, types.Integer)
@lower(operator.itruediv, types.NPTimedelta, types.Integer)
@lower(operator.floordiv, types.NPTimedelta, types.Integer)
@lower(operator.ifloordiv, types.NPTimedelta, types.Integer)
@lower(operator.truediv, types.NPTimedelta, types.Float)
@lower(operator.itruediv, types.NPTimedelta, types.Float)
@lower(operator.floordiv, types.NPTimedelta, types.Float)
@lower(operator.ifloordiv, types.NPTimedelta, types.Float)
def timedelta_over_number(context, builder, sig, args):
    td_arg, number_arg = args
    number_type = sig.args[1]
    ret = alloc_timedelta_result(builder)
    ok = builder.and_(
        is_not_nat(builder, td_arg),
        builder.not_(cgutils.is_scalar_zero_or_nan(builder, number_arg)),
    )
    with cgutils.if_likely(builder, ok):
        # Denominator is non-zero, non-NaN
        if isinstance(number_type, types.Float):
            val = builder.sitofp(td_arg, number_arg.type)
            val = builder.fdiv(val, number_arg)
            val = _cast_to_timedelta(context, builder, val)
        else:
            val = builder.sdiv(td_arg, number_arg)
        # The scaling is required for ufuncs np.*divide() with an explicit
        # output in a different unit.
        val = scale_timedelta(context, builder, val, sig.args[0], sig.return_type)
        builder.store(val, ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(operator.truediv, *TIMEDELTA_BINOP_SIG)
@lower(operator.itruediv, *TIMEDELTA_BINOP_SIG)
def timedelta_over_timedelta(context, builder, sig, args):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


@lower(operator.floordiv, *TIMEDELTA_BINOP_SIG)
def timedelta_floor_div_timedelta(context, builder, sig, args):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


def timedelta_mod_timedelta(context, builder, sig, args):
    # inspired by https://github.com/numpy/numpy/blob/fe8072a12d65e43bd2e0b0f9ad67ab0108cc54b3/numpy/core/src/umath/loops.c.src#L1424
    # alg is basically as `a % b`:
    # if a or b is NaT return NaT
    # elseif b is 0 return NaT
    # else pretend a and b are int and do pythonic int modulus

    raise NotImplementedError(_DEAD_CODEGEN_MSG)


# Comparison operators on timedelta64


def _create_timedelta_comparison_impl(ll_op, default_value):
    def impl(context, builder, sig, args):
        [va, vb] = args
        [ta, tb] = sig.args
        ret = alloc_boolean_result(builder)
        with builder.if_else(are_not_nat(builder, [va, vb])) as (
            then,
            otherwise,
        ):
            with then:
                try:
                    norm_a, norm_b = normalize_timedeltas(context, builder, va, vb, ta, tb)
                except RuntimeError:
                    # Cannot normalize units => the values are unequal (except if NaT)
                    builder.store(default_value, ret)
                else:
                    builder.store(builder.icmp_unsigned(ll_op, norm_a, norm_b), ret)
            with otherwise:
                # NaT ==/>=/>/</<= NaT is False
                # NaT != <anything, including NaT> is True
                if ll_op == "!=":
                    builder.store(cgutils.true_bit, ret)
                else:
                    builder.store(cgutils.false_bit, ret)
        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

    return impl


def _create_timedelta_ordering_impl(ll_op):
    def impl(context, builder, sig, args):
        [va, vb] = args
        [ta, tb] = sig.args
        ret = alloc_boolean_result(builder)
        with builder.if_else(are_not_nat(builder, [va, vb])) as (
            then,
            otherwise,
        ):
            with then:
                norm_a, norm_b = normalize_timedeltas(context, builder, va, vb, ta, tb)
                builder.store(builder.icmp_signed(ll_op, norm_a, norm_b), ret)
            with otherwise:
                # NaT >=/>/</<= NaT is False
                builder.store(cgutils.false_bit, ret)
        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

    return impl


timedelta_eq_timedelta_impl = _create_timedelta_comparison_impl("==", cgutils.false_bit)
timedelta_ne_timedelta_impl = _create_timedelta_comparison_impl("!=", cgutils.true_bit)
timedelta_lt_timedelta_impl = _create_timedelta_ordering_impl("<")
timedelta_le_timedelta_impl = _create_timedelta_ordering_impl("<=")
timedelta_gt_timedelta_impl = _create_timedelta_ordering_impl(">")
timedelta_ge_timedelta_impl = _create_timedelta_ordering_impl(">=")

for op_, func in [
    (operator.eq, timedelta_eq_timedelta_impl),
    (operator.ne, timedelta_ne_timedelta_impl),
    (operator.lt, timedelta_lt_timedelta_impl),
    (operator.le, timedelta_le_timedelta_impl),
    (operator.gt, timedelta_gt_timedelta_impl),
    (operator.ge, timedelta_ge_timedelta_impl),
]:
    lower(op_, *TIMEDELTA_BINOP_SIG)(func)


# Arithmetic on datetime64


def is_leap_year(builder, year_val):
    """
    Return a predicate indicating whether *year_val* (offset by 1970) is a
    leap year.
    """
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


def year_to_days(builder, year_val):
    """
    Given a year *year_val* (offset to 1970), return the number of days
    since the 1970 epoch.
    """
    # The algorithm below is copied from Numpy's get_datetimestruct_days()
    # (src/multiarray/datetime.c)
    ret = cgutils.alloca_once(builder, TIMEDELTA64)
    # First approximation
    days = scale_by_constant(builder, year_val, 365)
    # Adjust for leap years
    with builder.if_else(cgutils.is_neg_int(builder, year_val)) as (
        if_neg,
        if_pos,
    ):
        with if_pos:
            # At or after 1970:
            # 1968 is the closest leap year before 1970.
            # Exclude the current year, so add 1.
            from_1968 = add_constant(builder, year_val, 1)
            # Add one day for each 4 years
            p_days = builder.add(days, unscale_by_constant(builder, from_1968, 4))
            # 1900 is the closest previous year divisible by 100
            from_1900 = add_constant(builder, from_1968, 68)
            # Subtract one day for each 100 years
            p_days = builder.sub(p_days, unscale_by_constant(builder, from_1900, 100))
            # 1600 is the closest previous year divisible by 400
            from_1600 = add_constant(builder, from_1900, 300)
            # Add one day for each 400 years
            p_days = builder.add(p_days, unscale_by_constant(builder, from_1600, 400))
            builder.store(p_days, ret)
        with if_neg:
            # Before 1970:
            # NOTE `year_val` is negative, and so will be `from_1972` and `from_2000`.
            # 1972 is the closest later year after 1970.
            # Include the current year, so subtract 2.
            from_1972 = add_constant(builder, year_val, -2)
            # Subtract one day for each 4 years (`from_1972` is negative)
            n_days = builder.add(days, unscale_by_constant(builder, from_1972, 4))
            # 2000 is the closest later year divisible by 100
            from_2000 = add_constant(builder, from_1972, -28)
            # Add one day for each 100 years
            n_days = builder.sub(n_days, unscale_by_constant(builder, from_2000, 100))
            # 2000 is also the closest later year divisible by 400
            # Subtract one day for each 400 years
            n_days = builder.add(n_days, unscale_by_constant(builder, from_2000, 400))
            builder.store(n_days, ret)
    return builder.load(ret)


def reduce_datetime_for_unit(builder, dt_val, src_unit, dest_unit):
    dest_unit_code = npdatetime_helpers.DATETIME_UNITS[dest_unit]
    src_unit_code = npdatetime_helpers.DATETIME_UNITS[src_unit]
    if dest_unit_code < 2 or src_unit_code >= 2:
        return dt_val, src_unit
    # Need to compute the day ordinal for *dt_val*
    if src_unit_code == 0:
        # Years to days
        year_val = dt_val
        days_val = year_to_days(builder, year_val)

    else:
        # Months to days
        leap_array = cgutils.global_constant(builder, "leap_year_months_acc", leap_year_months_acc)
        normal_array = cgutils.global_constant(
            builder, "normal_year_months_acc", normal_year_months_acc
        )

        days = cgutils.alloca_once(builder, TIMEDELTA64)

        # First compute year number and month number
        year, month = cgutils.divmod_by_constant(builder, dt_val, 12)

        # Then deduce the number of days
        with builder.if_else(is_leap_year(builder, year)) as (then, otherwise):
            with then:
                addend = builder.load(cgutils.gep(builder, leap_array, 0, month, inbounds=True))
                builder.store(addend, days)
            with otherwise:
                addend = builder.load(cgutils.gep(builder, normal_array, 0, month, inbounds=True))
                builder.store(addend, days)

        days_val = year_to_days(builder, year)
        days_val = builder.add(days_val, builder.load(days))

    if dest_unit_code == 2:
        # Need to scale back to weeks
        weeks, _ = cgutils.divmod_by_constant(builder, days_val, 7)
        return weeks, "W"
    else:
        return days_val, "D"


def convert_datetime_for_arith(builder, dt_val, src_unit, dest_unit):
    """
    Convert datetime *dt_val* from *src_unit* to *dest_unit*.
    """
    # First partial conversion to days or weeks, if necessary.
    dt_val, dt_unit = reduce_datetime_for_unit(builder, dt_val, src_unit, dest_unit)
    # Then multiply by the remaining constant factor.
    dt_factor = npdatetime_helpers.get_timedelta_conversion_factor(dt_unit, dest_unit)
    if dt_factor is None:
        # This can happen when using explicit output in a ufunc.
        raise LoweringError("cannot convert datetime64 from %r to %r" % (src_unit, dest_unit))
    return scale_by_constant(builder, dt_val, dt_factor)


def _datetime_timedelta_arith(ll_op_name):
    def impl(context, builder, dt_arg, dt_unit, td_arg, td_unit, ret_unit):
        ret = alloc_timedelta_result(builder)
        with cgutils.if_likely(builder, are_not_nat(builder, [dt_arg, td_arg])):
            dt_arg = convert_datetime_for_arith(builder, dt_arg, dt_unit, ret_unit)
            td_factor = npdatetime_helpers.get_timedelta_conversion_factor(td_unit, ret_unit)
            td_arg = scale_by_constant(builder, td_arg, td_factor)
            ret_val = getattr(builder, ll_op_name)(dt_arg, td_arg)
            builder.store(ret_val, ret)
        return builder.load(ret)

    return impl


_datetime_plus_timedelta = _datetime_timedelta_arith("add")
_datetime_minus_timedelta = _datetime_timedelta_arith("sub")

# datetime64 + timedelta64


@lower(operator.add, types.NPDatetime, types.NPTimedelta)
@lower(operator.iadd, types.NPDatetime, types.NPTimedelta)
def datetime_plus_timedelta(context, builder, sig, args):
    dt_arg, td_arg = args
    dt_type, td_type = sig.args
    res = _datetime_plus_timedelta(
        context,
        builder,
        dt_arg,
        dt_type.unit,
        td_arg,
        td_type.unit,
        sig.return_type.unit,
    )
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(operator.add, types.NPTimedelta, types.NPDatetime)
@lower(operator.iadd, types.NPTimedelta, types.NPDatetime)
def timedelta_plus_datetime(context, builder, sig, args):
    td_arg, dt_arg = args
    td_type, dt_type = sig.args
    res = _datetime_plus_timedelta(
        context,
        builder,
        dt_arg,
        dt_type.unit,
        td_arg,
        td_type.unit,
        sig.return_type.unit,
    )
    return impl_ret_untracked(context, builder, sig.return_type, res)


# datetime64 - timedelta64


@lower(operator.sub, types.NPDatetime, types.NPTimedelta)
@lower(operator.isub, types.NPDatetime, types.NPTimedelta)
def datetime_minus_timedelta(context, builder, sig, args):
    dt_arg, td_arg = args
    dt_type, td_type = sig.args
    res = _datetime_minus_timedelta(
        context,
        builder,
        dt_arg,
        dt_type.unit,
        td_arg,
        td_type.unit,
        sig.return_type.unit,
    )
    return impl_ret_untracked(context, builder, sig.return_type, res)


# datetime64 - datetime64


@lower(operator.sub, types.NPDatetime, types.NPDatetime)
def datetime_minus_datetime(context, builder, sig, args):
    va, vb = args
    ta, tb = sig.args
    unit_a = ta.unit
    unit_b = tb.unit
    ret_unit = sig.return_type.unit
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, are_not_nat(builder, [va, vb])):
        va = convert_datetime_for_arith(builder, va, unit_a, ret_unit)
        vb = convert_datetime_for_arith(builder, vb, unit_b, ret_unit)
        ret_val = builder.sub(va, vb)
        builder.store(ret_val, ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


# datetime64 comparisons


def _create_datetime_comparison_impl(ll_op):
    def impl(context, builder, sig, args):
        va, vb = args
        ta, tb = sig.args
        unit_a = ta.unit
        unit_b = tb.unit
        ret_unit = npdatetime_helpers.get_best_unit(unit_a, unit_b)
        ret = alloc_boolean_result(builder)
        with builder.if_else(are_not_nat(builder, [va, vb])) as (
            then,
            otherwise,
        ):
            with then:
                norm_a = convert_datetime_for_arith(builder, va, unit_a, ret_unit)
                norm_b = convert_datetime_for_arith(builder, vb, unit_b, ret_unit)
                ret_val = builder.icmp_signed(ll_op, norm_a, norm_b)
                builder.store(ret_val, ret)
            with otherwise:
                if ll_op == "!=":
                    ret_val = cgutils.true_bit
                else:
                    ret_val = cgutils.false_bit
                builder.store(ret_val, ret)
        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

    return impl


datetime_eq_datetime_impl = _create_datetime_comparison_impl("==")
datetime_ne_datetime_impl = _create_datetime_comparison_impl("!=")
datetime_lt_datetime_impl = _create_datetime_comparison_impl("<")
datetime_le_datetime_impl = _create_datetime_comparison_impl("<=")
datetime_gt_datetime_impl = _create_datetime_comparison_impl(">")
datetime_ge_datetime_impl = _create_datetime_comparison_impl(">=")

for op, func in [
    (operator.eq, datetime_eq_datetime_impl),
    (operator.ne, datetime_ne_datetime_impl),
    (operator.lt, datetime_lt_datetime_impl),
    (operator.le, datetime_le_datetime_impl),
    (operator.gt, datetime_gt_datetime_impl),
    (operator.ge, datetime_ge_datetime_impl),
]:
    lower(op, *[types.NPDatetime] * 2)(func)


########################################################################
# datetime/timedelta fmax/fmin maximum/minimum support


def _gen_datetime_max_impl(NAT_DOMINATES):
    def datetime_max_impl(context, builder, sig, args):
        # note this could be optimizing relying on the actual value of NAT
        # but as NumPy doesn't rely on this, this seems more resilient
        in1, in2 = args
        in1_not_nat = is_not_nat(builder, in1)
        in2_not_nat = is_not_nat(builder, in2)
        in1_ge_in2 = builder.icmp_signed(">=", in1, in2)
        res = builder.select(in1_ge_in2, in1, in2)
        if NAT_DOMINATES:
            # NaT now dominates, like NaN
            in1, in2 = in2, in1
        res = builder.select(in1_not_nat, res, in2)
        res = builder.select(in2_not_nat, res, in1)

        return impl_ret_untracked(context, builder, sig.return_type, res)

    return datetime_max_impl


datetime_maximum_impl = _gen_datetime_max_impl(True)
datetime_fmax_impl = _gen_datetime_max_impl(False)


def _gen_datetime_min_impl(NAT_DOMINATES):
    def datetime_min_impl(context, builder, sig, args):
        # note this could be optimizing relying on the actual value of NAT
        # but as NumPy doesn't rely on this, this seems more resilient
        in1, in2 = args
        in1_not_nat = is_not_nat(builder, in1)
        in2_not_nat = is_not_nat(builder, in2)
        in1_le_in2 = builder.icmp_signed("<=", in1, in2)
        res = builder.select(in1_le_in2, in1, in2)
        if NAT_DOMINATES:
            # NaT now dominates, like NaN
            in1, in2 = in2, in1
        res = builder.select(in1_not_nat, res, in2)
        res = builder.select(in2_not_nat, res, in1)

        return impl_ret_untracked(context, builder, sig.return_type, res)

    return datetime_min_impl


datetime_minimum_impl = _gen_datetime_min_impl(True)
datetime_fmin_impl = _gen_datetime_min_impl(False)


def _gen_timedelta_max_impl(NAT_DOMINATES):
    def timedelta_max_impl(context, builder, sig, args):
        # note this could be optimizing relying on the actual value of NAT
        # but as NumPy doesn't rely on this, this seems more resilient
        in1, in2 = args
        in1_not_nat = is_not_nat(builder, in1)
        in2_not_nat = is_not_nat(builder, in2)
        in1_ge_in2 = builder.icmp_signed(">=", in1, in2)
        res = builder.select(in1_ge_in2, in1, in2)
        if NAT_DOMINATES:
            # NaT now dominates, like NaN
            in1, in2 = in2, in1
        res = builder.select(in1_not_nat, res, in2)
        res = builder.select(in2_not_nat, res, in1)

        return impl_ret_untracked(context, builder, sig.return_type, res)

    return timedelta_max_impl


timedelta_maximum_impl = _gen_timedelta_max_impl(True)
timedelta_fmax_impl = _gen_timedelta_max_impl(False)


def _gen_timedelta_min_impl(NAT_DOMINATES):
    def timedelta_min_impl(context, builder, sig, args):
        # note this could be optimizing relying on the actual value of NAT
        # but as NumPy doesn't rely on this, this seems more resilient
        in1, in2 = args
        in1_not_nat = is_not_nat(builder, in1)
        in2_not_nat = is_not_nat(builder, in2)
        in1_le_in2 = builder.icmp_signed("<=", in1, in2)
        res = builder.select(in1_le_in2, in1, in2)
        if NAT_DOMINATES:
            # NaT now dominates, like NaN
            in1, in2 = in2, in1
        res = builder.select(in1_not_nat, res, in2)
        res = builder.select(in2_not_nat, res, in1)

        return impl_ret_untracked(context, builder, sig.return_type, res)

    return timedelta_min_impl


timedelta_minimum_impl = _gen_timedelta_min_impl(True)
timedelta_fmin_impl = _gen_timedelta_min_impl(False)


def _cast_to_timedelta(context, builder, val):
    temp = builder.alloca(TIMEDELTA64)
    val_is_nan = builder.fcmp_unordered("uno", val, val)
    with builder.if_else(val_is_nan) as (then, els):
        with then:
            # NaN does not guarantee to cast to NAT.
            # We should store NAT explicitly.
            builder.store(NAT, temp)
        with els:
            builder.store(builder.fptosi(val, TIMEDELTA64), temp)
    return builder.load(temp)


@lower_cast(types.NPDatetime, types.Integer)
@lower_cast(types.NPTimedelta, types.Integer)
def _cast_npdatetime_int64(context, builder, fromty, toty, val):
    if toty.bitwidth != 64:  # all date time types are 64 bit
        msg = f"Cannot cast {fromty} to {toty} as {toty} is not 64 bits wide."
        raise ValueError(msg)
    return val


@overload_method(types.NPTimedelta, "__hash__")
@overload_method(types.NPDatetime, "__hash__")
def ol_hash_npdatetime(x):
    if numpy_support.numpy_version >= (2, 2) and isinstance(x, types.NPTimedelta) and not x.unit:
        raise ValueError("Can't hash generic timedelta64")

    if IS_32BITS:

        def impl(x):
            x = np.int64(x)
            if x < 2**31 - 1:  # x < LONG_MAX
                y = np.int32(x)
            else:
                hi = (np.int64(x) & 0xFFFFFFFF00000000) >> 32
                lo = np.int64(x) & 0x00000000FFFFFFFF
                y = np.int32(lo + (1000003) * hi)
            if y == -1:
                y = np.int32(-2)
            return y

    else:

        def impl(x):
            if np.int64(x) == -1:
                return np.int64(-2)
            return np.int64(x)

    return impl


lower(npdatetime_helpers.datetime_minimum, types.NPDatetime, types.NPDatetime)(
    datetime_minimum_impl
)
lower(npdatetime_helpers.datetime_minimum, types.NPTimedelta, types.NPTimedelta)(
    datetime_minimum_impl
)
lower(npdatetime_helpers.datetime_maximum, types.NPDatetime, types.NPDatetime)(
    datetime_maximum_impl
)
lower(npdatetime_helpers.datetime_maximum, types.NPTimedelta, types.NPTimedelta)(
    datetime_maximum_impl
)




# Autogenerated by mlir-tblgen; don't manually edit.

from enum import IntEnum, auto, IntFlag
from ._ods_common import _cext as _ods_cext
from ..ir import register_attribute_builder
_ods_ir = _ods_cext.ir

class BinaryFn(IntEnum):
    """allowed 32-bit signless integer cases: 0, 1, 2, 3, 4, 5, 6, 7, 8, 9"""

    add = 0
    sub = 1
    mul = 2
    div = 3
    div_unsigned = 4
    max_signed = 5
    min_signed = 6
    max_unsigned = 7
    min_unsigned = 8
    powf = 9

    def __str__(self):
        if self is BinaryFn.add:
            return "add"
        if self is BinaryFn.sub:
            return "sub"
        if self is BinaryFn.mul:
            return "mul"
        if self is BinaryFn.div:
            return "div"
        if self is BinaryFn.div_unsigned:
            return "div_unsigned"
        if self is BinaryFn.max_signed:
            return "max_signed"
        if self is BinaryFn.min_signed:
            return "min_signed"
        if self is BinaryFn.max_unsigned:
            return "max_unsigned"
        if self is BinaryFn.min_unsigned:
            return "min_unsigned"
        if self is BinaryFn.powf:
            return "powf"
        raise ValueError("Unknown BinaryFn enum entry.")



@register_attribute_builder("BinaryFn", allow_existing=True)
def _binaryfn(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class ElementwiseArityGroup(IntEnum):
    """allowed 32-bit signless integer cases: 1, 2, 3"""

    Unary = 1
    Binary = 2
    Ternary = 3

    def __str__(self):
        if self is ElementwiseArityGroup.Unary:
            return "Unary"
        if self is ElementwiseArityGroup.Binary:
            return "Binary"
        if self is ElementwiseArityGroup.Ternary:
            return "Ternary"
        raise ValueError("Unknown ElementwiseArityGroup enum entry.")



@register_attribute_builder("ElementwiseArityGroup", allow_existing=True)
def _elementwisearitygroup(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class ElementwiseCaseLimits(IntEnum):
    """allowed 32-bit signless integer cases: """

    LastUnary = 13
    LastBinary = 23
    LastTernary = 24

    def __str__(self):
        if self is ElementwiseCaseLimits.LastUnary:
            return "LastUnary"
        if self is ElementwiseCaseLimits.LastBinary:
            return "LastBinary"
        if self is ElementwiseCaseLimits.LastTernary:
            return "LastTernary"
        raise ValueError("Unknown ElementwiseCaseLimits enum entry.")



@register_attribute_builder("ElementwiseCaseLimits", allow_existing=True)
def _elementwisecaselimits(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class ElementwiseKind(IntEnum):
    """allowed 32-bit signless integer cases: 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23"""

    exp = 0
    log = 1
    abs = 2
    ceil = 3
    floor = 4
    negf = 5
    reciprocal = 6
    round = 7
    sqrt = 8
    rsqrt = 9
    square = 10
    tanh = 11
    erf = 12
    add = 13
    sub = 14
    mul = 15
    div = 16
    div_unsigned = 17
    max_signed = 18
    min_signed = 19
    max_unsigned = 20
    min_unsigned = 21
    powf = 22
    select = 23

    def __str__(self):
        if self is ElementwiseKind.exp:
            return "exp"
        if self is ElementwiseKind.log:
            return "log"
        if self is ElementwiseKind.abs:
            return "abs"
        if self is ElementwiseKind.ceil:
            return "ceil"
        if self is ElementwiseKind.floor:
            return "floor"
        if self is ElementwiseKind.negf:
            return "negf"
        if self is ElementwiseKind.reciprocal:
            return "reciprocal"
        if self is ElementwiseKind.round:
            return "round"
        if self is ElementwiseKind.sqrt:
            return "sqrt"
        if self is ElementwiseKind.rsqrt:
            return "rsqrt"
        if self is ElementwiseKind.square:
            return "square"
        if self is ElementwiseKind.tanh:
            return "tanh"
        if self is ElementwiseKind.erf:
            return "erf"
        if self is ElementwiseKind.add:
            return "add"
        if self is ElementwiseKind.sub:
            return "sub"
        if self is ElementwiseKind.mul:
            return "mul"
        if self is ElementwiseKind.div:
            return "div"
        if self is ElementwiseKind.div_unsigned:
            return "div_unsigned"
        if self is ElementwiseKind.max_signed:
            return "max_signed"
        if self is ElementwiseKind.min_signed:
            return "min_signed"
        if self is ElementwiseKind.max_unsigned:
            return "max_unsigned"
        if self is ElementwiseKind.min_unsigned:
            return "min_unsigned"
        if self is ElementwiseKind.powf:
            return "powf"
        if self is ElementwiseKind.select:
            return "select"
        raise ValueError("Unknown ElementwiseKind enum entry.")



@register_attribute_builder("ElementwiseKind", allow_existing=True)
def _elementwisekind(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class IteratorType(IntEnum):
    """Iterator type"""

    parallel = 0
    reduction = 1

    def __str__(self):
        if self is IteratorType.parallel:
            return "parallel"
        if self is IteratorType.reduction:
            return "reduction"
        raise ValueError("Unknown IteratorType enum entry.")



@register_attribute_builder("IteratorType", allow_existing=True)
def _iteratortype(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class TernaryFn(IntEnum):
    """allowed 32-bit signless integer cases: 0"""

    select = 0

    def __str__(self):
        if self is TernaryFn.select:
            return "select"
        raise ValueError("Unknown TernaryFn enum entry.")



@register_attribute_builder("TernaryFn", allow_existing=True)
def _ternaryfn(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class TypeFn(IntEnum):
    """allowed 32-bit signless integer cases: 0, 1"""

    cast_signed = 0
    cast_unsigned = 1

    def __str__(self):
        if self is TypeFn.cast_signed:
            return "cast_signed"
        if self is TypeFn.cast_unsigned:
            return "cast_unsigned"
        raise ValueError("Unknown TypeFn enum entry.")



@register_attribute_builder("TypeFn", allow_existing=True)
def _typefn(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class UnaryFn(IntEnum):
    """allowed 32-bit signless integer cases: 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12"""

    exp = 0
    log = 1
    abs = 2
    ceil = 3
    floor = 4
    negf = 5
    reciprocal = 6
    round = 7
    sqrt = 8
    rsqrt = 9
    square = 10
    tanh = 11
    erf = 12

    def __str__(self):
        if self is UnaryFn.exp:
            return "exp"
        if self is UnaryFn.log:
            return "log"
        if self is UnaryFn.abs:
            return "abs"
        if self is UnaryFn.ceil:
            return "ceil"
        if self is UnaryFn.floor:
            return "floor"
        if self is UnaryFn.negf:
            return "negf"
        if self is UnaryFn.reciprocal:
            return "reciprocal"
        if self is UnaryFn.round:
            return "round"
        if self is UnaryFn.sqrt:
            return "sqrt"
        if self is UnaryFn.rsqrt:
            return "rsqrt"
        if self is UnaryFn.square:
            return "square"
        if self is UnaryFn.tanh:
            return "tanh"
        if self is UnaryFn.erf:
            return "erf"
        raise ValueError("Unknown UnaryFn enum entry.")



@register_attribute_builder("UnaryFn", allow_existing=True)
def _unaryfn(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class WinogradConv2DFmr(IntEnum):
    """allowed 32-bit signless integer cases: 0, 1, 2"""

    F_2_3 = 0
    F_4_3 = 1
    F_2_5 = 2

    def __str__(self):
        if self is WinogradConv2DFmr.F_2_3:
            return "F_2_3"
        if self is WinogradConv2DFmr.F_4_3:
            return "F_4_3"
        if self is WinogradConv2DFmr.F_2_5:
            return "F_2_5"
        raise ValueError("Unknown WinogradConv2DFmr enum entry.")



@register_attribute_builder("WinogradConv2DFmr", allow_existing=True)
def _winogradconv2dfmr(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

@register_attribute_builder("linalg.BinaryFnAttr")
def _binaryfnattr(x, context):
    return _ods_ir.Attribute.parse(f'#linalg.binary_fn<{str(x)}>', context=context)

@register_attribute_builder("linalg.ElementwiseKindAttr")
def _elementwisekindattr(x, context):
    return _ods_ir.Attribute.parse(f'#linalg.elementwise_kind<{str(x)}>', context=context)

@register_attribute_builder("linalg.IteratorTypeEnum")
def _iteratortypeenum(x, context):
    return _ods_ir.Attribute.parse(f'#linalg.iterator_type<{str(x)}>', context=context)

@register_attribute_builder("linalg.TernaryFnAttr")
def _ternaryfnattr(x, context):
    return _ods_ir.Attribute.parse(f'#linalg.ternary_fn<{str(x)}>', context=context)

@register_attribute_builder("linalg.TypeFnAttr")
def _typefnattr(x, context):
    return _ods_ir.Attribute.parse(f'#linalg.type_fn<{str(x)}>', context=context)

@register_attribute_builder("linalg.UnaryFnAttr")
def _unaryfnattr(x, context):
    return _ods_ir.Attribute.parse(f'#linalg.unary_fn<{str(x)}>', context=context)




# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
CUDA vector type stubs (float32x4, int32x2, etc.)

These are user-facing objects that can be used in CUDA kernels to construct
vector types. They map to numba_cuda_mlir's VectorType internally.
"""

import itertools
import numpy as np

from numba_cuda_mlir.type_defs.vector_types import VectorType
from numba_cuda_mlir import types

BASE_TYPE_MAP = {
    "int8": types.int8,
    "int16": types.int16,
    "int32": types.int32,
    "int64": types.int64,
    "uint8": types.uint8,
    "uint16": types.uint16,
    "uint32": types.uint32,
    "uint64": types.uint64,
    "float16": types.float16,
    "float32": types.float32,
    "float64": types.float64,
}


def make_vector_types():
    """Create instances for all CUDA vector types."""
    vector_types_by_name = {}
    vector_type_prefix = (
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "float16",
        "float32",
        "float64",
    )
    vector_type_element_counts = (1, 2, 3, 4)

    for prefix, nelem in itertools.product(vector_type_prefix, vector_type_element_counts):
        base_type = BASE_TYPE_MAP[prefix]
        vec_type = VectorType(base_type, nelem)
        vector_types_by_name[vec_type.name] = vec_type

    return vector_types_by_name


def make_vector_type_aliases(vector_types_by_name):
    """Create C-compatible aliases for vector types (e.g., float4 -> float32x4)."""
    base_type_to_alias = {
        "char": f"int{np.dtype(np.byte).itemsize * 8}",
        "short": f"int{np.dtype(np.short).itemsize * 8}",
        "int": f"int{np.dtype(np.intc).itemsize * 8}",
        "long": f"int{np.dtype(np.int_).itemsize * 8}",
        "longlong": f"int{np.dtype(np.longlong).itemsize * 8}",
        "uchar": f"uint{np.dtype(np.ubyte).itemsize * 8}",
        "ushort": f"uint{np.dtype(np.ushort).itemsize * 8}",
        "uint": f"uint{np.dtype(np.uintc).itemsize * 8}",
        "ulong": f"uint{np.dtype(np.uint).itemsize * 8}",
        "ulonglong": f"uint{np.dtype(np.ulonglong).itemsize * 8}",
        "half": "float16",
        "float": f"float{np.dtype(np.single).itemsize * 8}",
        "double": f"float{np.dtype(np.double).itemsize * 8}",
    }

    vector_types_by_alias = {}
    for alias_prefix, base_type_prefix in base_type_to_alias.items():
        for nelem in (1, 2, 3, 4):
            alias_name = f"{alias_prefix}{nelem}"
            target_name = f"{base_type_prefix}x{nelem}"
            if target_name in vector_types_by_name:
                vector_types_by_alias[alias_name] = vector_types_by_name[target_name]

    return vector_types_by_alias


# Build lookup dictionaries
vector_types_by_name = make_vector_types()
vector_types_by_alias = make_vector_type_aliases(vector_types_by_name)

# List of all unique vector types
_vector_types = list(vector_types_by_name.values())

# Export all types as module-level attributes
for name, vec_type in vector_types_by_name.items():
    globals()[name] = vec_type
for alias, vec_type in vector_types_by_alias.items():
    globals()[alias] = vec_type

# List of all exported names
__all__ = list(vector_types_by_name.keys()) + list(vector_types_by_alias.keys())



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from .._transform_pdl_extension_ops_gen import *
from .._transform_pdl_extension_ops_gen import _Dialect

try:
    from ...ir import *
    from .._ods_common import (
        get_op_result_or_value as _get_op_result_or_value,
        get_op_results_or_values as _get_op_results_or_values,
        _cext as _ods_cext,
    )
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from typing import Union


@_ods_cext.register_operation(_Dialect, replace=True)
class PDLMatchOp(PDLMatchOp):
    def __init__(
        self,
        result_type: Type,
        target: Union[Operation, Value],
        pattern_name: Union[Attribute, str],
        *,
        loc=None,
        ip=None,
    ):
        super().__init__(
            result_type,
            _get_op_result_or_value(target),
            pattern_name,
            loc=loc,
            ip=ip,
        )


@_ods_cext.register_operation(_Dialect, replace=True)
class WithPDLPatternsOp(WithPDLPatternsOp):
    def __init__(self, target: Union[Operation, Value, Type], *, loc=None, ip=None):
        root = _get_op_result_or_value(target) if not isinstance(target, Type) else None
        root_type = target if isinstance(target, Type) else root.type
        super().__init__(root=root, loc=loc, ip=ip)
        self.regions[0].blocks.append(root_type)

    @property
    def body(self) -> Block:
        return self.regions[0].blocks[0]

    @property
    def bodyTarget(self) -> Value:
        return self.body.arguments[0]
